#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from importlib.metadata import version

__version__ = version("appligator")


__all__ = [
    "__version__",
]



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from pathlib import Path
from typing import Annotated

import typer

EOZILLA_PATH = Path(__file__).parent.parent.parent.parent.resolve()
DEFAULT_DAGS_FOLDER = EOZILLA_PATH / "eozilla-airflow/dags"
PROCESS_REGISTRY_SPEC_EX = "wraptile.services.local.testing:service.process_registry"

CLI_NAME = "appligator"

cli = typer.Typer(name=CLI_NAME)


@cli.command()
def main(
    process_registry_spec: Annotated[
        str | None,
        typer.Argument(
            ...,
            help=f"Process registry specification. For example {PROCESS_REGISTRY_SPEC_EX!r}.",
        ),
    ] = None,
    dags_folder: Annotated[
        Path,
        typer.Option(..., help="An Airflow DAGs folder to which to write the outputs."),
    ] = DEFAULT_DAGS_FOLDER,
    version: Annotated[
        bool,
        typer.Option(..., help="Show version and exit."),
    ] = False,
):
    """
    Generate various application formats from your processing workflows.

    WARNING: This tool is under development and subject to change anytime.

    Currently it expects a _process registry_ as input, which must be
    provided in form a Python module path plus an attribute path separated
    by a colon: "my.module.path:my.registry_obj". The type of the registry
    must be `procodile.ProcessRegistry`. In the future the tool will be
    able to handle other input types.

    It is also currently limited to generating DAGs for Airflow 3+.
    The plan is to extend it to also output Docker images with or
    without metadata such as the OGC CWL standard (= EOAP).
    """
    import datetime

    from appligator import __version__
    from appligator.airflow.gen_dag import gen_dag
    from gavicore.util.dynimp import import_value
    from procodile import ProcessRegistry

    if version:
        typer.echo(f"{__version__}")
        raise typer.Exit(0)

    if not process_registry_spec:
        typer.echo("Error: missing process registry specification.")
        raise typer.Exit(1)

    process_registry: ProcessRegistry = import_value(
        process_registry_spec,
        type=ProcessRegistry,
        name="process_registry",
        example=PROCESS_REGISTRY_SPEC_EX,
    )
    dags_folder.mkdir(exist_ok=True)
    for process_id, process in process_registry.items():
        dag_code = gen_dag(process)
        dag_file = dags_folder / f"{process_id}.py"
        with dag_file.open("w") as stream:
            stream.write(
                f"# WARNING - THIS IS GENERATED CODE\n"
                f"#   Generator: Eozilla Appligator v{__version__}\n"
                f"#        Date: {datetime.datetime.now().isoformat()}\n"
                f"\n"
                f"{dag_code}"
            )


if __name__ == "__main__":  # pragma: no cover
    cli()



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from typing import Any

from gavicore.models import InputDescription, ProcessDescription
from procodile import Process


def gen_dag(process: Process | ProcessDescription, function_module: str = "") -> str:
    if isinstance(process, Process):
        process_description = process.description
        # noinspection PyUnresolvedReferences
        function_module = function_module or process.function.__module__
    elif isinstance(process, ProcessDescription):
        process_description = process
        if not function_module:
            raise ValueError("function_module argument is required")
    else:
        raise TypeError(f"unexpected type for process: {type(process).__name__}")

    return _gen_dag(process_description, function_module)


def _gen_dag(process_description: ProcessDescription, function_module: str) -> str:
    function_name = process_description.id
    input_descriptions = process_description.inputs or {}

    param_specs = [
        f"{param_name!r}: Param({get_param_args(input_description)})"
        for param_name, input_description in input_descriptions.items()
    ]

    tab = "    "
    num_outputs = len(process_description.outputs or [])
    lines = [
        "from airflow.sdk import Param, dag, task",
        "",
        f"from {function_module} import {function_name}",
        "",
        "",
        "@dag(",
        f"{tab}{function_name!r},",
        f"{tab}dag_display_name={process_description.title!r},",
        f"{tab}description={process_description.description!r},",
        f"{tab}params=" + "{",
        *[f"{tab}{tab}{p}," for p in param_specs],
        f"{tab}" + "},",
        f"{tab}is_paused_upon_creation=False,",
        ")",
        f"def {function_name}_dag():",
        "",
        f"{tab}@task(multiple_outputs={(num_outputs > 1)!r})",
        f"{tab}def {function_name}_task(params):",
        f"{tab}{tab}return {function_name}(**params)",
        "",
        f"{tab}task_instance = {function_name}_task()  # noqa: F841",
        "",
        f"{function_name}_dag()",
    ]
    return "\n".join(lines) + "\n"


def get_param_args(input_description: InputDescription):
    schema = dict(
        input_description.schema_.model_dump(
            mode="json",
            by_alias=True,
            exclude_defaults=True,
            exclude_none=True,
            exclude_unset=True,
        )
    )
    param_args: list[tuple[str, Any]] = []
    if "default" in schema:
        param_args.append(("default", schema.pop("default")))
    if "type" in schema:
        param_args.append(("type", schema.pop("type")))
    if input_description.title:
        schema.pop("title", None)
        param_args.append(("title", input_description.title))
    if input_description.description:
        schema.pop("description", None)
        param_args.append(("description", input_description.description))
    param_args.extend(sorted(schema.items(), key=lambda item: item[0]))
    return ", ".join(f"{sk}={sv!r}" for sk, sv in param_args)
